// bc_smg.cpp
// ============================================
// 
// This surfacemap generator was created for use
// with BC42.  It provides a smoother texture, which of course
// everybody likes.  It isn't perfect however, as it is still
// completely based on the grid-formed material map.  There
// will always be issues like this until a method which does
// not use the material map is implemented...
// 
// Copyright (c)2003 by Digital Illusions CE AB. All Rights Reserved.

/*
MATERIAL MAP DATA BY TYPE
	0 Default
	1 Water/Ice/Slippery
	2 Dry Grass
	3 Wet Grass
	4 Dry Dirt
	5 Wet Dirt
	6 Mud
	7 Outside Map (Can't go there or you will be shot)
	8 Gravel
	9 Frozen Ground
	10 Dry Sand
	11 Wet Sand
	12 Rock Surface
	13 Sand Road
	14 Dirt Road
	15 Paved Road
*/


#include "stdafx.h"
#include "resource.h"
#include <malloc.h>
#include <stdlib.h>
#include <math.h>

#include "bc_smg.h"



// DllMain()
// ============================================
//
// The main entry point of the DLL.  This is called when the DLL is
// loaded and is called when BattleCraft 1942 first starts up.
//
BOOL APIENTRY DllMain(HINSTANCE hInst, DWORD  reason_for_call, LPVOID lpReserved)
{
	hInstance = hInst;
    return TRUE;
}



// DialogWndProc
// ============================================
//
// Callback for Dialog box
//
LRESULT CALLBACK DialogWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
			{
				return TRUE;
			}
		case WM_DESTROY:
			{
				EndDialog(hDlg, IDOK);
				return TRUE;
			}

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
			case IDOK:
				{
					EndDialog(hDlg, IDOK);
					return TRUE;
				}
			case IDCANCEL:
				{
					EndDialog(hDlg, IDCANCEL);
					return TRUE;
				}
			}
			break;
	}
    return FALSE;
}



// initSurfaceMap()
// ============================================
//
// This is called when the user selects the "Generate Surface Maps" option
// from within BC42.  Return TRUE to continue the generation process or return
// FALSE and BC42 will abort the process.
// 
// matmap = a pointer to the material map data stored in a multidimensional array
// max = the max array size of the map (256, 512, 1024)
// offx, offy = Map offsets
//
extern "C" __declspec(dllexport) bool initSurfaceMap(matmap_s *original_matmap, int original_mapsize, int original_x_offset, int original_y_offset)
{
	// Tell user what they are using
	int ret = DialogBox(hInstance, (LPCTSTR)IDD_SMG_DIALOG, NULL, (DLGPROC)DialogWndProc);
	if(ret != IDOK)
		return false;

	// Get some variables passed from BC42
	mapsize = original_mapsize;
	texture_x_offset = original_x_offset;
	texture_y_offset = original_y_offset;

	// Make a local copy of the material map (hmm, maybe just use the pointer)
	memcpy(material_map, original_matmap, (1033*1033)*sizeof(matmap_s));

	// Edit surrounding materials so that the texture wraps seamlessly around outside edges
	for(int i=0; i<mapsize+2; i++)
	{
		material_map[i][mapsize] = material_map[i][0];
		material_map[mapsize+1][i] = material_map[1][i];
	}

	return true;
}



// initSurfaceBmp()
// ============================================
//
// BC42 calls this to send a copy of the loaded bitmap TextureSet
// specified by the user
//
// index = the number of the surfacebmp array to store
// surfbmp = the surface bmp data as stored by BC42
//
extern "C" __declspec(dllexport) void initSurfaceBmp(int index, surfacebmps_s *surfbmp)
{
	memcpy(&surfacebmps[index], surfbmp, sizeof(surfacebmps_s));
}



// generateSurfaceMap()
// ============================================
//
// This is called for each terrain texture
//
// size = The size of the map to generate.  BC42 always passes 1024.
// tx = The X map grid coordinate.
// ty = The Y map grid coordinate.
//
extern "C" __declspec(dllexport) unsigned char *generateSurfaceMap(int size, int tx, int ty)
{
	// Allocate the memory needed to store our surface map
	// BC42 will turn this into a .DDS file and store it in the level
	// RFA upon return. (1024x1024 24-bit image file)
	unsigned char *final_texture = (unsigned char*)malloc(size*size*3);

	// Get the name of the DLL module called (?)
	GetModuleFileName(NULL, (LPTSTR)module, MAXMODULE);

	// Some loop variables
	const int num_materials = 4;

	matmap_s materials[num_materials];
	surfacebmps_s textures[num_materials];
	float material_blend[num_materials], total;
	int index, z;
	unsigned char original_r[num_materials], original_g[num_materials], original_b[num_materials], final_red, final_green, final_blue;

	//apply the texture offsets
	tx += texture_x_offset;
	ty += texture_y_offset;

	// Generate texture coordinates TxTXxTY
	int i_start = tx*64;
	int i_end = tx*64 + 64;
	int j_start = ty*64;
	int j_end = ty*64 + 64;
	for(int i=i_start; i<i_end; i++)
	{
		for(int j=j_start; j<j_end; j++)
		{
			// Begin detail loops for this heightmap coordinate
			for(int x=0; x<16; x++)
			{
				for(int y=0; y<16; y++)
				{
					// Get surrounding materials
					materials[0] = material_map[mapsize-i+0][j+0];
					materials[1] = material_map[mapsize-i+1][j+0];
					materials[2] = material_map[mapsize-i+1][j+1];
					materials[3] = material_map[mapsize-i+0][j+1];

					// Calculate Blend Factors
					material_blend[0] = (18.0f - sqrt(pow(x, 2) + pow(y, 2))) / 15.0f;
					material_blend[1] = (18.0f - sqrt(pow(15-x, 2) + pow(y, 2))) / 15.0f;
					material_blend[2] = (18.0f - sqrt(pow(15-x, 2) + pow(15-y, 2))) / 15.0f;
					material_blend[3] = (18.0f - sqrt(pow(x, 2) + pow(15-y, 2))) / 15.0f;

					// Check Blend Factors
					for(z=0; z<num_materials; z++)
					{
						if(material_blend[z] > 1.0f)
							material_blend[z] = 1.0f;
						else if(material_blend[z] < 0.0f)
							material_blend[z] = 0.0f;
					}

					// Normalize Blend Factors
					total = 0.0f;
					for(z=0; z<num_materials; z++)
						total += material_blend[z];
					for(z=0; z<num_materials; z++)
						material_blend[z] /= total;


					// Get surrounding textures
					for(z=0; z<num_materials; z++)
					{
						textures[z] = surfacebmps[materials[z].type];
					}

					// Get texture bitmap index (256x256x3 bytes, wrapped)
					index = ((((i-i_start)*16)%256)*3)+((((j-j_start)*16)%256)*3*256)+((15-x)*3)+(y*3*256);

					// Get Texture Color Information
					for(z=0; z<num_materials; z++)
					{
						original_r[z] = textures[z].data[index+0];
						original_g[z] = textures[z].data[index+1];
						original_b[z] = textures[z].data[index+2];
					}


					// Find final colors
					final_red = 0;
					final_green = 0;
					final_blue = 0;
					for(z=0; z<num_materials; z++)
					{
						final_red	+= original_r[z] * material_blend[z];
						final_green	+= original_g[z] * material_blend[z];
						final_blue	+= original_b[z] * material_blend[z];
					}

					// Get final bitmap index (1024x1024x3 bytes)
					index = ((i-i_start)*16*3)+((j-j_start)*16*3*1024)+((15-x)*3)+(y*3*1024);

					// Set Color for this index
					final_texture[index+0] = final_red;
					final_texture[index+1] = final_green;
					final_texture[index+2] = final_blue;
				}
			}
		}
	}

	// Return our surface map back to BC42
	return final_texture;
}
